#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>

#include "devdlg.h"

#include <pr.h>

/* Defined in winutil.c */
extern HWND hWndMain;
extern HINSTANCE  ghInstance;    

PR_DWORD PickedDriver = 0;
PR_DWORD PickedDevice = 0;
static HWND devicebox = NULL;

extern void mdeinit(void);
extern void PR_ShutDown (void);

extern LPDIRECTDRAW wgtpdd1;



static HRESULT WINAPI D3D_EnumDeviceFunc
                        (LPGUID lpGuid, LPSTR lpDeviceDescription,
                        LPSTR lpDeviceName, LPD3DDEVICEDESC lpHWDesc,
                        LPD3DDEVICEDESC lpHELDesc, LPVOID lpContext)
{
  /* Record the D3D driver's inforamation */
	if(devicebox)
	SendMessage (devicebox, CB_ADDSTRING, 0, (LPARAM)lpDeviceName);
	
	lstrcpy(PR_D3D_Drivers[PR_D3D_NumDrivers].Name, lpDeviceName);
	PR_D3D_NumDrivers++;
	
	return (D3DENUMRET_OK);
}




void EnumerateDevices (int driver)
{
HRESULT ddrval; 

  if (driver != PR_PrimaryDriver)
    ddrval = DirectDrawCreate (&PR_DDDeviceList[driver], &wgtpdd1, NULL);
  else
    ddrval = DirectDrawCreate (NULL, &wgtpdd1, NULL);
 
  /* create the main DirectDraw object */
  if (ddrval != DD_OK)
    {
     if (diagnostics_level & 2)
       wgt_log_message ("File: %s\nLine: %d\nError: %s\n", __FILE__, __LINE__, getErrorString(ddrval));
     PR_FatalError ("DirectDrawCreate Failed", "init");
    }
 
  
  ddrval= IDirectDraw_QueryInterface (wgtpdd1, &IID_IDirectDraw2, (LPVOID *)&wgtpdd);
  if (ddrval != DD_OK)
    {
     if (diagnostics_level & 2)
       wgt_log_message ("File: %s\nLine: %d\nError: %s\n", __FILE__, __LINE__, getErrorString(ddrval));
     PR_FatalError ("Query Interface Failed", "init");
    }


  /* Now create the Direct3D interface */
  ddrval= IDirectDraw_QueryInterface (wgtpdd, &IID_IDirect3D2, (LPVOID *)&wgtd3d);
  if (ddrval != DD_OK)
    {
     if (diagnostics_level & 2)
       wgt_log_message ("File: %s\nLine: %d\nError: %s\n", __FILE__, __LINE__, PR_getErrorString(ddrval));
     PR_FatalError ("Query Interface Failed", "PR_DetectD3D");
    }

  PR_D3D_NumDrivers = 0;
  ddrval = IDirect3D2_EnumDevices(wgtd3d,
                                  D3D_EnumDeviceFunc, NULL);
  if (ddrval != DD_OK)
    {
     if (diagnostics_level & 2)
       wgt_log_message ("File: %s\nLine: %d\nError: %s\n", __FILE__, __LINE__, PR_getErrorString(ddrval));
     PR_FatalError ("Enum Devices Failed", "PR_DetectD3D");
    }



  /* Now we can get rid of all the interfaces */
  ddrval = wgtd3d->lpVtbl->Release(wgtd3d);
  wgtd3ddev = NULL;

  ddrval = wgtpdd->lpVtbl->Release(wgtpdd);
  wgtpdd = NULL;

  ddrval = wgtpdd1->lpVtbl->Release(wgtpdd1);
  wgtpdd1 = NULL;
}






void FillDeviceBox (int driver, HWND box)
{
  while (ComboBox_GetCount (box) > 0)
    ComboBox_DeleteString (box, 0);

  EnumerateDevices (driver);

  ComboBox_SetCurSel (box, 0);
}







BOOL CALLBACK DevDlgProc (HWND hDlg, UINT Message, WPARAM wParam, LPARAM lParam)
{
HWND driverbox;
int i;

RECT rect;
int w, h;

  switch (Message)
   {
    case WM_INITDIALOG:
		// center dialog
		GetWindowRect(hDlg, &rect);
		w = rect.right - rect.left;
		h = rect.bottom - rect.top;

		rect.left = (GetSystemMetrics(SM_CXFULLSCREEN)-w) >> 1;
		rect.top = (GetSystemMetrics(SM_CYFULLSCREEN)-h) >> 1;
		SetWindowPos(hDlg, HWND_TOPMOST, rect.left, rect.top, w, h, SWP_SHOWWINDOW);


      driverbox = GetDlgItem (hDlg, IDC_DRIVERCOMBO);

      for (i = 0; i < PR_DD_NumDrivers; i++)
        SendMessage (driverbox, CB_ADDSTRING, 0, (LPARAM)PR_DDDeviceNames[i]);
          ComboBox_SetCurSel (driverbox, 0);
      devicebox = GetDlgItem (hDlg, IDC_DEVCOMBO);
      FillDeviceBox (0, devicebox);

      return TRUE;
          break;

    case WM_COMMAND:
      if ((LOWORD(wParam) == IDC_DRIVERCOMBO) && 
                  (HIWORD(wParam) == CBN_SELCHANGE))
        {
         driverbox = GetDlgItem (hDlg, IDC_DRIVERCOMBO);
         devicebox = GetDlgItem (hDlg, IDC_DEVCOMBO);
         PickedDriver = ComboBox_GetCurSel (driverbox);
             FillDeviceBox (PickedDriver, devicebox);
         return TRUE;
        }

      if ((LOWORD(wParam) == IDC_DEVCOMBO) && 
                  (HIWORD(wParam) == CBN_SELCHANGE))
        {
         devicebox = GetDlgItem (hDlg, IDC_DEVCOMBO);
         PickedDevice = ComboBox_GetCurSel (devicebox);
             return TRUE;
        }
 
      if (LOWORD(wParam) == IDOK)
        {
         EndDialog (hDlg, LOWORD(wParam));

         if (Button_GetCheck (GetDlgItem(hDlg, IDC_CHECK1)))
           PR_Settings.Use32Bit = 1;
         else
           PR_Settings.Use32Bit = 0;

         if (Button_GetCheck (GetDlgItem(hDlg, IDC_CHECK2)))
           PR_Settings.UseAGP = 1;
         else
           PR_Settings.UseAGP = 0;

         if (Button_GetCheck (GetDlgItem(hDlg, IDC_CHECK3)))
           PR_Settings.UseTrilinearFiltering = 1;
         else
           PR_Settings.UseTrilinearFiltering = 0;

         if (Button_GetCheck (GetDlgItem(hDlg, IDC_CHECK4)))
           PR_Settings.UseTripleBuffering = 1;
         else
           PR_Settings.UseTripleBuffering = 0;

         return TRUE;
        }

      if (LOWORD(wParam) == IDCANCEL)
        {
         EndDialog (hDlg, LOWORD(wParam));
         PR_ShutDown ();
         mdeinit ();
         DestroyWindow(hWndMain);


         return FALSE;
        }
        
   }
  return FALSE;
}


void PR_ChooseDDraw_Driver (int *driver, int *device)
	typedef struct CFG_HEADER
	{
		int driver, device;
	};
{
// modified by DarrenNix
// Try to load .cfg file. If it isn't there or is corrupt,
// get info from user and then try creating the file.

	HANDLE hInFile, hOutFile;
	DWORD dwNumWrit, dwNumRead;
	struct CFG_HEADER cfghdr;

	hInFile = CreateFile("video.cfg", GENERIC_READ, FILE_SHARE_READ,
		NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hInFile != INVALID_HANDLE_VALUE)
	{
		// cfg file exists
		do
		{
			ReadFile(hInFile, &cfghdr, sizeof(cfghdr), &dwNumRead, NULL);
			if(dwNumRead != sizeof(cfghdr)) break; // file's corrupt

			PickedDriver = *driver = cfghdr.driver;
			PickedDevice = *device = cfghdr.device;
			
			EnumerateDevices(*driver);

			// load settings from file
			ReadFile(hInFile, &PR_Settings, sizeof(PR_Settings), &dwNumRead, NULL);
			if(dwNumRead != sizeof(PR_Settings)) break; // file's corrupt

			// loaded successfuly
			CloseHandle(hInFile);
			return;
		} while(0);

		// close corrupt file
		CloseHandle(hInFile);
	}

	// No cfg file or it's corrupt, so ask user what they want:
	if(IDOK != DialogBox (ghInstance, MAKEINTRESOURCE (IDD_DEVDIALOG), hWndMain, (DLGPROC)DevDlgProc))
		return; // will ask next time
	
	*driver = PickedDriver;
	*device = PickedDevice;

	// Contents in saving must match loading...
	cfghdr.driver = *driver;
	cfghdr.device = *device;

	// write new cfg file:
	hOutFile = CreateFile("video.cfg", GENERIC_WRITE, FILE_SHARE_READ,
		NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	if(hOutFile == INVALID_HANDLE_VALUE)
		return; // will ask again next time

	do
	{
		WriteFile(hOutFile, &cfghdr, sizeof(cfghdr), &dwNumWrit, NULL);
		if(dwNumWrit != sizeof(cfghdr)) break;

		WriteFile(hOutFile, &PR_Settings, sizeof(PR_Settings), &dwNumWrit, NULL);
		if(dwNumWrit != sizeof(PR_Settings)) break;

	} while(0);

	CloseHandle(hOutFile);
	return;
}
