#include <conio.h>
#include <stdlib.h>

#include <stdarg.h>

#include <pr.h>
#include <prgui.h>

#ifdef __3DFX__
#include <pr3dfx.h>
#include <glide.h>
#endif

#ifdef WIN32
#include <windows.h>
#include <winutil.h>
#endif

// my includes
#include "camera.h"
#include "trigtables.h"

CCamera::CCamera()
{
	Init();
}

CCamera::~CCamera()
{
	Release();
}

void CCamera::Create(PR_DWORD dwCreateMode/*=CAMFLAG_AIM_TARGET*/)
{
	m_PRCamera = PR_AllocCamera();
	PR_InitializeCamera(m_PRCamera);
	PR_SetCameraMode(m_PRCamera, dwCreateMode);

	// this initial view yields the optimizations for
	// looking around. To set a different initial view,
	// you should change the angles m_nPsi, or m_nTheta.
	PR_PositionCameraSource (m_PRCamera, 0.0f, 0.0f, 0.0f);
	PR_PositionCameraTarget (m_PRCamera, 0.0f, 0.0f, 1.0f);
	
	// By remembering the angle between the actual,
	// and this initial view, it is possible to calculate
	// the actual view from those angles.

	m_nIsValid = 1;
}

void CCamera::Activate(bool bSetView/*=true*/)
{
	// changes to camera will only be available
	// after this function is called. Should be
	// called once per frame.

	if(bSetView)
	{
		PR_REAL x, y, z;
		x = m_PRCamera->source.x + SIN[m_nPsi]*COS[m_nTheta]; // derived from rotational matrices
		y = m_PRCamera->source.y + -SIN[m_nTheta];			 // optimized from initial left handed view
		z = m_PRCamera->source.z + COS[m_nPsi]*COS[m_nTheta];
		PR_PositionCameraTarget(m_PRCamera, x, y, z);
	}

	PR_SetActiveCamera(m_PRCamera);
}

void CCamera::Release()
{
	if(IsValid()) PR_FreeCamera(m_PRCamera);
	Init();
}

void CCamera::HandleMouseLook(bool bInvLook/*=false*/)
{
	// NOTE: This will affect other cameras if there
	//       are multiple cameras utilizing the mouse
	//       look feature via this function.

	int dx = mouse.mx - 160, // change in pos
		dy = mouse.my - 100;

	mouse.mx = 160; // reset position
	mouse.my = 100;

	// if the terminology is incorrect,
	// that's because I made it up.

	HandleMouseLook(dx, dy, bInvLook);
}

void CCamera::HandleMouseLook(int dx, int dy, bool bInvLook/*=false*/)
{
	// turn body left or right
	m_nPsi += dx;

	// make sure they can look all around
	if(m_nPsi > TRIGTABLE_MAXINDEX) m_nPsi = 0;
	else if(m_nPsi < 0) m_nPsi = TRIGTABLE_MAXINDEX;

	// tilt head up or down
	m_nTheta += (bInvLook ? (-dy) : (dy));

	// 0 is strait, so this is necessary
	if(m_nTheta < 0) m_nTheta = TRIGTABLE_MAXINDEX;
	else if(m_nTheta > TRIGTABLE_MAXINDEX) m_nTheta = 0;

	// make sure they can't look too far up or down
	if(m_nTheta < 410 && m_nTheta > 160) m_nTheta = 160; // strait down
	else if(m_nTheta > 410 && m_nTheta < 600) m_nTheta = 600; // strait up
	// 410 is strait back from where they are moving

	// the new view direction is calculated in
	// Activate, since it is called only once
	// per frame.
}

void CCamera::Init()
{
	m_PRCamera = NULL;
	m_nTheta = m_nPsi = 0;
	m_nIsValid = 0;
}

void CCamera::GetAngle(int &nTheta, int &nPsi)
{
	nTheta = m_nTheta;
	nPsi = m_nPsi;
}

void CCamera::OnMove(PR_REAL fDelta)
{
	PR_CameraDirection(m_PRCamera); // must prepare internals
	PR_DollyCamera(m_PRCamera, fDelta);
	// remember to call Activate to finalize settings
}

void CCamera::OnStraffe(PR_REAL fDelta)
{
	// Positive is to the right, negative
	// is to the left. As one would expect.
	fDelta = -fDelta;

	PR_CameraDirection(m_PRCamera); // must prepare internals

	PR_POINT
		right,
		&up = m_PRCamera->up,
		&dir = m_PRCamera->direction;

	PR_VecCrossProduct(right, up, dir);

	PR_REAL
		dx = right.x*fDelta,
		dy = right.y*fDelta,
		dz = right.z*fDelta;

	MovePos(dx, dy, dz);
	MoveTarget(dx, dy, dz);

	// remember to call Activate
	// to finalize settings
}

void CCamera::SetPos(PR_REAL x, PR_REAL y, PR_REAL z)
{
	PR_PositionCameraSource(m_PRCamera, x, y, z);
}

void CCamera::SetTarget(PR_REAL x, PR_REAL y, PR_REAL z)
{
	PR_PositionCameraTarget(m_PRCamera, x, y, z);
}

void CCamera::MovePos(PR_REAL dx, PR_REAL dy, PR_REAL dz)
{
	PR_MoveCameraSource(m_PRCamera, dx, dy, dz);
}

void CCamera::MoveTarget(PR_REAL dx, PR_REAL dy, PR_REAL dz)
{
	PR_MoveCameraTarget(m_PRCamera, dx, dy, dz);
}

PR_CAMERA* CCamera::GetPRCamera()
{
	return m_PRCamera;
}

void CCamera::GetPos(PR_REAL &x, PR_REAL &y, PR_REAL &z)
{
	x = m_PRCamera->source.x;
	y = m_PRCamera->source.y;
	z = m_PRCamera->source.z;
}

void CCamera::GetTarget(PR_REAL &x, PR_REAL &y, PR_REAL &z)
{
	x = m_PRCamera->dest.x;
	y = m_PRCamera->dest.y;
	z = m_PRCamera->dest.z;
}

void CCamera::Create(PR_CAMERA &prc, PR_DWORD dwCreateMode/*=CAMFLAG_AIM_TARGET*/)
{
	m_PRCamera = PR_AllocCamera();
	PR_InitializeCamera(m_PRCamera);
	PR_SetCameraMode(m_PRCamera, dwCreateMode);
	PR_PositionCameraSource (m_PRCamera, 0.0f, 0.0f, 0.0f);
	PR_PositionCameraTarget (m_PRCamera, 0.0f, 0.0f, 1.0f);
	
	// This function inherits most of the settings from the prc
	// parameter, but not the orientation. If you need to inherit
	// the orientation, you must calculate the m_nTheta, and
	// m_nPsi variables and copy the prc members associated
	// with orientation into the m_PRCamera structure.

	m_PRCamera->aspect_ratio = prc.aspect_ratio;
	m_PRCamera->tilt_angle = prc.tilt_angle;
	m_PRCamera->distance = prc.distance;
	m_PRCamera->flags = prc.flags;
	m_PRCamera->fov = prc.fov;
	m_PRCamera->nearclip = prc.nearclip;
	m_PRCamera->farclip = prc.farclip;
	m_PRCamera->num_frames = prc.num_frames;
	m_PRCamera->num_cam_keys = prc.num_cam_keys;
	m_PRCamera->cam_keys = new PRKF_CAMKEY[prc.num_cam_keys];
	memcpy(m_PRCamera->cam_keys, prc.cam_keys, sizeof(PRKF_CAMKEY) * prc.num_cam_keys);

	m_nIsValid = 1;
}
