#include <conio.h>
#include <stdlib.h>

#include <stdarg.h>

#include <pr.h>
#include <prgui.h>

#ifdef __3DFX__
#include <pr3dfx.h>
#include <glide.h>
#endif

#ifdef WIN32
#include <windows.h>
#include <winutil.h>
#endif

#include "lights.h"

///////////////////////////////////////////////////////////////////////
// CLight

	// public operations //
CLight::CLight()
{
	Init();
}

CLight::~CLight()
{
	Init();
}

void CLight::SetPos(PR_REAL x, PR_REAL y, PR_REAL z)
{
	PR_SetLightPosition(m_pPRLightList, m_nIndex, x, y, z);
}

void CLight::SetColor(PR_REAL r, PR_REAL g, PR_REAL b)
{
	PR_SetLightColor(m_pPRLightList, m_nIndex, r, g, b);
}

void CLight::SetColor(int r, int g, int b)
{
	PR_SetLightColor(m_pPRLightList, m_nIndex, r/255.0f, g/255.0f, b/255.0f);
}

void CLight::SetStrength(PR_REAL fIntensity)
{
	PR_SetLightStrength(m_pPRLightList, m_nIndex, fIntensity);
}

void CLight::TurnOn()
{
	PR_SetLightOn(m_pPRLightList, m_nIndex);
}

void CLight::TurnOff()
{
	PR_SetLightOff(m_pPRLightList, m_nIndex);
}

	// protected operations //
void CLight::Init()
{
	m_nIndex = -1;
	m_pPRLightList = NULL;
}

///////////////////////////////////////////////////////////////////////
// CLightList

	// public operations //
CLightList::CLightList()
{
	Init();
}

CLightList::~CLightList()
{
	Release();
}

void CLightList::Release()
{
	// user can clear light list and reuse
	if(m_pLights) delete [] m_pLights;
	Init();
}

bool CLightList::Create(int nMaxLights)
{
	PR_AllocLights (&m_PRLightList, nMaxLights);

	m_pLights = new CLight[nMaxLights];
	if(!m_pLights) return false;
	
	m_PRLightList.NumLights = 0;
	m_nMaxLights = nMaxLights;

	// user must CreateLight

	return true;
}

CLight* CLightList::CreateLight(PR_DWORD dwLightType/*=DIRECTIONAL_LIGHT*/, PR_DWORD dwLightFalloff/*=200*/)
{
	if(m_nNumLights >= m_nMaxLights) return NULL;

	CLight *pRet = &m_pLights[m_nNumLights];
	pRet->m_nIndex = m_nNumLights;
	pRet->m_pPRLightList = &m_PRLightList;
	
	PR_SetLightType(&m_PRLightList, m_nNumLights, dwLightType);
	PR_SetLightStrength(&m_PRLightList, m_nNumLights, 1.0);

	if(dwLightType == POINT_LIGHT)
	{
		// this depends on the world's scale
		PR_SetLightFalloff(&m_PRLightList, m_nNumLights, dwLightFalloff);
	}

	m_nNumLights++;
	m_PRLightList.NumLights = m_nNumLights;

	return pRet;
}

void CLightList::DoTransform()
{
	// this function must be called every frame
    PR_TransformLights(&m_PRLightList);
}

	// protected operations //
void CLightList::Init()
{
	m_pLights = NULL;
	m_nNumLights = m_nMaxLights = 0;
}

CLight* CLightList::GetLight(UINT nIndex)
{
	if(nIndex >= m_nNumLights)
		return NULL;

	return &m_pLights[nIndex];
}
