// main.h - darrennix@msn.com
// http://homepages.msn.com/RedmondAve/darrennix
// See also: www.egerter.com www.gamedev.net
//////////////////////////////////////////////////////////////////////

// The usual disclaimers apply to all source code you receive
// from me. You can read my copyright/disclaimer at my website.

#include <Windows.h>
#include <time.h>
#include <stdio.h>

// My includes
#include <CDirectDraw.h>
#include <CBmpFile.h>

#include <CAviFile.h>

// Globals ///////////////////////////////////////////////////////////

#define SCREEN_WIDTH 640
#define SCREEN_HEIGHT 480

#define		CLASS_NAME		"AVITest Class"
#define		TITLE_STRING	"AVITest"
HWND		g_hWnd			=NULL;
HINSTANCE	g_hInstance		=NULL;
CDirectDraw g_dd;
DWORD		g_bAppActive	=TRUE;

CAviFile	g_AviFile;
CBmpFile	g_BmpFile;

// Debugging //////////////////////////////////////////////////////////

#ifdef _DEBUG

#ifdef VERIFY
#undef VERIFY
#define VERIFY(B) \
	if(!(B))\
	{\
		MessageBeep(MB_ICONSTOP);\
		g_dd.Release();\
		char szMsg[256];\
		sprintf(szMsg, "Fatal Error in file %s, line %d. Failed to verify \"%s\"",\
			__FILE__, __LINE__, #B);\
		PostQuitMessage(1);\
	}
#endif

#else

#ifdef VERIFY
#undef VERIFY
#define VERIFY(B)
#endif

#endif

///////////////////////////////////////////////////////////////////////

long FAR PASCAL WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_DESTROY:
		ShowCursor(TRUE);
		PostQuitMessage(0);
		break;
	case WM_KEYDOWN:
		switch(wParam)
		{
		case VK_ESCAPE:
			g_bAppActive = false;
			break;
		}
		break;
	case WM_ACTIVATEAPP:
		g_bAppActive = wParam;
		break;
	case WM_QUIT:
		g_bAppActive = false;
		break;
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
    
    return 0L;
}

BOOL OnRegisterClass()
{
	WNDCLASS wndClass;
    
	wndClass.style              = CS_HREDRAW | CS_VREDRAW;
    wndClass.lpfnWndProc        = WndProc;
    wndClass.cbClsExtra         = 0;
    wndClass.cbWndExtra         = 0;
    wndClass.hInstance          = g_hInstance;
    wndClass.hIcon              = LoadIcon(g_hInstance, "IDI_APPICON");
    wndClass.hCursor            = LoadCursor(NULL, IDC_ARROW);
    wndClass.hbrBackground      = (HBRUSH) GetStockObject(BLACK_BRUSH);
    wndClass.lpszMenuName       = NULL;
    wndClass.lpszClassName      = CLASS_NAME;

	return RegisterClass(&wndClass);
}

HWND OnCreateWindow()
{
    return CreateWindowEx(WS_EX_APPWINDOW, CLASS_NAME, TITLE_STRING,
		WS_POPUP | WS_SYSMENU, 0, 0, GetSystemMetrics(SM_CXSCREEN),
		GetSystemMetrics(SM_CYSCREEN), NULL, NULL, g_hInstance, NULL);
}

int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	// Go to black...
	g_hInstance = hInstance;
	if(!OnRegisterClass()) return 1;
	if(g_hWnd = OnCreateWindow(), !g_hWnd) return 1;
	ShowWindow(g_hWnd, nCmdShow);
	UpdateWindow(g_hWnd);
	ShowCursor(FALSE);

	// Open AVI...
	VERIFY(g_AviFile.Open("klowner.avi"));
	if(g_AviFile.GetVideoStreamCount() == 0)
		return -1;

	// Open background image...
	VERIFY(g_BmpFile.Load("sphere.bmp"));
	
	// Start video retrieve... (this file only has the one stream)
	VERIFY(g_AviFile.StartVideoRetrieve(0));

	// Change display mode...
	g_dd.Create(g_hWnd, SCREEN_WIDTH, SCREEN_HEIGHT, 16);

	// Prepare main loop's data...
	MSG msg;
	memset(&msg, 0, sizeof(MSG));

	LPBITMAPINFOHEADER pBmp;			// Decompressed video frame

	LONG
		lFrameTime=0,					// Millisecs into video clip
		lNewTime=0,						// Current time
		lOldTime=timeGetTime(),			// Time at last drawn frame
		lElapsedTime=0,					// Time between drawn frames
		lVideoDirection=1;				// 1=Forward -1=Reverse

	LPBITMAPINFO pVideoFormat =			// AVI File's video format
		g_AviFile.GetVideoFormat(0);

	// Frame surface
	LPDIRECTDRAWSURFACE pFrameSurface = // I'll use ddraw to render
		g_dd.CreateSurface(pVideoFormat->bmiHeader.biWidth, pVideoFormat->bmiHeader.biHeight);

	// (calculate frame's dimensions)
	RECT rcDest;						// Destination rect of video
	
	{ // limit scope for temporary variables
		int nHalfWidth = (pVideoFormat->bmiHeader.biWidth >> 1),
			nHalfHeight = (pVideoFormat->bmiHeader.biHeight >> 1);
		
		rcDest.left = SCREEN_WIDTH/2 - nHalfWidth;
		rcDest.top = SCREEN_HEIGHT/2 - nHalfHeight;
		rcDest.right = SCREEN_WIDTH/2 + nHalfWidth;
		rcDest.bottom = SCREEN_HEIGHT/2 + nHalfHeight;
	}

	// Background image surface
	LPDIRECTDRAWSURFACE pBkgSurface =
		g_dd.CreateSurface(g_BmpFile);
	VERIFY(pBkgSurface);

	// (calculate background image's dimensions, and velocity)
	RECT rcBkgDest;
	rcBkgDest.left = rcBkgDest.top = 0;

	int bkg_dx = 2, bkg_dy = 3,
		bkg_width = g_BmpFile.Width(),
		bkg_height = g_BmpFile.Height();

	// Enter main loop...
	while(g_bAppActive)
	{
		// Synchronize frame play to real time...
		lNewTime = timeGetTime();
		lElapsedTime = lNewTime - lOldTime;
		lFrameTime += lElapsedTime * lVideoDirection;

		while(PeekMessage(&msg, g_hWnd, 0, 0, TRUE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		g_AviFile.GetVideoFrameAtTime(0, lFrameTime, &pBmp);
		if(!pBmp)
		{
			// no more frames, so change the playback
			// direction.
			lVideoDirection = -lVideoDirection;
			continue;
		}
	
		g_dd.ClearSurface(g_dd.GetBackSurface());

		// Copy frame to IDirectDrawSurface object...
		g_dd.BlitToSurface(pFrameSurface, (LPBITMAPINFO)pBmp);

		// UNCOMMENT THIS BLOCK IF YOU WANT TO SEE THE TRANSPARANCY,
		// I DIDN'T LIKE THE WAY IT LOOKED, SO I COMMENTED IT OUT.
		/*// Blit background image to back buffer...
		g_dd.Blit(&rcBkgDest, pBkgSurface);
		
		// translate background image
		rcBkgDest.left += bkg_dx;
		rcBkgDest.top += bkg_dy;
		rcBkgDest.right = rcBkgDest.left + bkg_width;
		rcBkgDest.bottom = rcBkgDest.top + bkg_height;
		if(rcBkgDest.left > SCREEN_WIDTH - bkg_width || rcBkgDest.left < 0)
			bkg_dx = -bkg_dx;
		if(rcBkgDest.top > SCREEN_HEIGHT - bkg_height || rcBkgDest.top < 0)
			bkg_dy = -bkg_dy;*/

		// Anything blitted here would be behind the
		// AVI sequence.
		
		// Blit frame to back buffer...
		g_dd.Blit(&rcDest, pFrameSurface);

		// Anything blitted here would be on top of the
		// AVI sequence.

		lOldTime = lNewTime;

		g_dd.Flip();
	}

destroy:
	g_AviFile.EndVideoRetrieve(0);
	g_AviFile.Release();

	// g_BmpFile destroyed automatically

	pBkgSurface->Release();
	pFrameSurface->Release();
	g_dd.Release();

	return 0;
}