// CDirectDraw.cpp: implementation of the CDirectDraw class.
//
//////////////////////////////////////////////////////////////////////

#define DIRECTDRAW_VERSION 0x0500.

#include <Windows.h>
#include <ddraw.h>
#include <nassert.h>
#include "CDirectDraw.h"

// Defines...
#ifdef ProcessResult
#undef ProcessResult
#endif

#define ProcessResult(H) {\
	HRESULT hResult = (H);\
	if(hResult!=DD_OK) ProcessResultA(hResult, __FILE__, __LINE__);\
	}

// module structures
struct CALLBACK_PARAM
{
	DWORD dwDesiredCaps;
	LPDIRECTDRAW *ppDD;
};

typedef CALLBACK_PARAM *LPCALLBACK_PARAM;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CDirectDraw::CDirectDraw()
{
	Init();
}

CDirectDraw::~CDirectDraw()
{
	Release();
}

bool CDirectDraw::CreateFor3D(HWND hWnd, int nWidth, int nHeight, int nBitDepth)
{
	int nSuccess = Create(hWnd, nWidth, nHeight, nBitDepth, DDSCAPS_3DDEVICE);

	if(!nSuccess)
	{ ASSERT(FALSE); return false; }

	ASSERT(m_lpDD);
	ASSERT(m_lpBackSurface);

	// nSuccess > 0 ? hardware acclerated : software emulation
	if(!CreateZBuffer(nSuccess > 0, nWidth, nHeight)) return false;
	
	return true;
}

void CDirectDraw::Init()
{
	m_lpBackSurface = NULL;
	m_lpDD = NULL;
	m_lpPrimarySurface = NULL;
	m_lpZSurface = NULL;
	m_nBpp = 0;
}

void CDirectDraw::Release()
{
	RELEASE(m_lpZSurface);
	RELEASE(m_lpBackSurface);
	RELEASE(m_lpPrimarySurface);
	RELEASE(m_lpDD);
}

void CDirectDraw::ProcessResultA(HRESULT hResult, LPSTR lpszFile, int nLine)
{
	char szMessage[128];

	switch(hResult)
	{
	case DDERR_ALREADYINITIALIZED:
		strcpy(szMessage, "The object has already been initialized.");
		break;
	case DDERR_BLTFASTCANTCLIP:
		strcpy(szMessage, "A DirectDrawClipper object is attached to a source surface that has passed into a call to the IDirectDrawSurface3::BltFast method.");
		break;
	case DDERR_CANNOTATTACHSURFACE:
		strcpy(szMessage, "A surface cannot be attached to another requested surface.");
		break;
	case DDERR_CANNOTDETACHSURFACE:
		strcpy(szMessage, "A surface cannot be detached from another requested surface.");
		break;
	case DDERR_CANTCREATEDC:
		strcpy(szMessage, "Windows cannot create any more device contexts (DCs).");
		break;
	case DDERR_CANTDUPLICATE:
		strcpy(szMessage, "Primary and 3-D surfaces, or surfaces that are implicitly created, cannot be duplicated.");
		break;
	case DDERR_CANTLOCKSURFACE:
		strcpy(szMessage, "Access to this surface is refused because an attempt was made to lock the primary surface without DCI support.");
		break;
	case DDERR_CANTPAGELOCK:
		strcpy(szMessage, "An attempt to page lock a surface failed. Page lock will not work on a display-memory surface or an emulated primary surface.");
		break;
	case DDERR_CANTPAGEUNLOCK:
		strcpy(szMessage, "An attempt to page unlock a surface failed. Page unlock will not work on a display-memory surface or an emulated primary surface.");
		break;
	case DDERR_CLIPPERISUSINGHWND:
		strcpy(szMessage, "An attempt was made to set a clip list for a DirectDrawClipper object that is already monitoring a window handle.");
		break;
	case DDERR_COLORKEYNOTSET:
		strcpy(szMessage, "No source color key is specified for this operation.");
		break;
	case DDERR_CURRENTLYNOTAVAIL:
		strcpy(szMessage, "No support is currently available.");
		break;
	case DDERR_DCALREADYCREATED:
		strcpy(szMessage, "A device context (DC) has already been returned for this surface. Only one DC can be retrieved for each surface.");
		break;
	case DDERR_DEVICEDOESNTOWNSURFACE:
		strcpy(szMessage, "Surfaces created by one direct draw device cannot be used directly by another direct draw device");
		break;
	case DDERR_DIRECTDRAWALREADYCREATED:
		strcpy(szMessage, "A DirectDraw object representing this driver has already been created for this process.");
		break;
	case DDERR_EXCEPTION:
		strcpy(szMessage, "An exception was encountered while performing the requested operation.");
		break;
	case DDERR_EXCLUSIVEMODEALREADYSET:
		strcpy(szMessage, "An attempt was made to set the cooperative level when it was already set to exclusive.");
		break;
	case DDERR_GENERIC:
		strcpy(szMessage, "There is an undefined error condition.");
		break;
	case DDERR_HEIGHTALIGN:
		strcpy(szMessage, "The height of the provided rectangle is not a multiple of the required alignment.");
		break;
	case DDERR_HWNDALREADYSET:
		strcpy(szMessage, "The DirectDraw cooperative level window handle has already been set. It cannot be reset while the process has surfaces or palettes created.");
		break;
	case DDERR_HWNDSUBCLASSED:
		strcpy(szMessage, "DirectDraw is prevented from restoring state because the DirectDraw cooperative level window handle has been subclassed.");
		break;
	case DDERR_IMPLICITLYCREATED:
		strcpy(szMessage, "The surface cannot be restored because it is an implicitly created surface.");
		break;
	case DDERR_INCOMPATIBLEPRIMARY:
		strcpy(szMessage, "The primary surface creation request does not match with the existing primary surface.");
		break;
	case DDERR_INVALIDCAPS:
		strcpy(szMessage, "One or more of the capability bits passed to the callback function are incorrect.");
		break;
	case DDERR_INVALIDCLIPLIST:
		strcpy(szMessage, "DirectDraw does not support the provided clip list.");
		break;
	case DDERR_INVALIDDIRECTDRAWGUID:
		strcpy(szMessage, "The globally unique identifier (GUID) passed to the DirectDrawCreate function is not a valid DirectDraw driver identifier.");
		break;
	case DDERR_INVALIDMODE:
		strcpy(szMessage, "DirectDraw does not support the requested mode.");
		break;
	case DDERR_INVALIDOBJECT:
		strcpy(szMessage, "DirectDraw received a pointer that was an invalid DirectDraw object.");
		break;
	case DDERR_INVALIDPARAMS:
		strcpy(szMessage, "One or more of the parameters passed to the method are incorrect.");
		break;
	case DDERR_INVALIDPIXELFORMAT:
		strcpy(szMessage, "The pixel format was invalid as specified.");
		break;
	case DDERR_INVALIDPOSITION:
		strcpy(szMessage, "The position of the overlay on the destination is no longer legal.");
		break;
	case DDERR_INVALIDRECT:
		strcpy(szMessage, "The provided rectangle was invalid.");
		break;
	case DDERR_INVALIDSURFACETYPE:
		strcpy(szMessage, "The requested operation could not be performed because the surface was of the wrong type.");
		break;
	case DDERR_LOCKEDSURFACES:
		strcpy(szMessage, "One or more surfaces are locked, causing the failure of the requested operation.");
		break;
	case DDERR_MOREDATA:
		strcpy(szMessage, "There is more data available than the specified buffer size can hold");
		break;
	case DDERR_NO3D:
		strcpy(szMessage, "No 3-D hardware or emulation is present.");
		break;
	case DDERR_NOALPHAHW:
		strcpy(szMessage, "No alpha acceleration hardware is present or available, causing the failure of the requested operation.");
		break;
	case DDERR_NOBLTHW:
		strcpy(szMessage, "No blitter hardware is present.");
		break;
	case DDERR_NOCLIPLIST:
		strcpy(szMessage, "No clip list is available.");
		break;
	case DDERR_NOCLIPPERATTACHED:
		strcpy(szMessage, "No DirectDrawClipper object is attached to the surface object.");
		break;
	case DDERR_NOCOLORCONVHW:
		strcpy(szMessage, "The operation cannot be carried out because no color-conversion hardware is present or available.");
		break;
	case DDERR_NOCOLORKEY:
		strcpy(szMessage, "The surface does not currently have a color key.");
		break;
	case DDERR_NOCOLORKEYHW:
		strcpy(szMessage, "The operation cannot be carried out because there is no hardware support for the destination color key.");
		break;
	case DDERR_NOCOOPERATIVELEVELSET:
		strcpy(szMessage, "A create function is called without the IDirectDraw2::SetCooperativeLevel method being called.");
		break;
	case DDERR_NODC:
		strcpy(szMessage, "No DC has ever been created for this surface.");
		break;
	case DDERR_NODDROPSHW:
		strcpy(szMessage, "No DirectDraw raster operation (ROP) hardware is available.");
		break;
	case DDERR_NODIRECTDRAWHW:
		strcpy(szMessage, "Hardware-only DirectDraw object creation is not possible; the driver does not support any hardware.");
		break;
	case DDERR_NODIRECTDRAWSUPPORT:
		strcpy(szMessage, "DirectDraw support is not possible with the current display driver.");
		break;
	case DDERR_NOEMULATION:
		strcpy(szMessage, "Software emulation is not available.");
		break;
	case DDERR_NOEXCLUSIVEMODE:
		strcpy(szMessage, "The operation requires the application to have exclusive mode, but the application does not have exclusive mode.");
		break;
	case DDERR_NOFLIPHW:
		strcpy(szMessage, "Flipping visible surfaces is not supported.");
		break;
	case DDERR_NOFOCUSWINDOW:
		strcpy(szMessage, "An attempt was made to create or set a device window without first setting the focus window");
		break;
	case DDERR_NOGDI:
		strcpy(szMessage, "No GDI is present.");
		break;
	case DDERR_NOHWND:
		strcpy(szMessage, "Clipper notification requires a window handle, or no window handle has been previously set as the cooperative level window handle.");
		break;
	case DDERR_NOMIPMAPHW:
		strcpy(szMessage, "The operation cannot be carried out because no mipmap texture mapping hardware is present or available.");
		break;
	case DDERR_NOMIRRORHW:
		strcpy(szMessage, "The operation cannot be carried out because no mirroring hardware is present or available.");
		break;
	case DDERR_NONONLOCALVIDMEM:
		strcpy(szMessage, "An attempt was made to allocate non-local video memory from a device that does not support non-local video memory");
		break;
	case DDERR_NOOPTIMIZEHW:
		strcpy(szMessage, "The device does not support optimized surfaces");
		break;
	case DDERR_NOOVERLAYDEST:
		strcpy(szMessage, "The IDirectDrawSurface3::GetOverlayPosition method is called on an overlay that the IDirectDrawSurface3::UpdateOverlay method has not been called on to establish a destination.");
		break;
	case DDERR_NOOVERLAYHW:
		strcpy(szMessage, "The operation cannot be carried out because no overlay hardware is present or available.");
		break;
	case DDERR_NOPALETTEATTACHED:
		strcpy(szMessage, "No palette object is attached to this surface.");
		break;
	case DDERR_NOPALETTEHW:
		strcpy(szMessage, "There is no hardware support for 16- or 256-color palettes.");
		break;
	case DDERR_NORASTEROPHW:
		strcpy(szMessage, "The operation cannot be carried out because no appropriate raster operation hardware is present or available.");
		break;
	case DDERR_NOROTATIONHW:
		strcpy(szMessage, "The operation cannot be carried out because no rotation hardware is present or available.");
		break;
	case DDERR_NOSTRETCHHW:
		strcpy(szMessage, "The operation cannot be carried out because there is no hardware support for stretching.");
		break;
	case DDERR_NOT4BITCOLOR:
		strcpy(szMessage, "The DirectDrawSurface object is not using a 4-bit color palette and the requested operation requires a 4-bit color palette.");
		break;
	case DDERR_NOT4BITCOLORINDEX:
		strcpy(szMessage, "The DirectDrawSurface object is not using a 4-bit color index palette and the requested operation requires a 4-bit color index palette.");
		break;
	case DDERR_NOT8BITCOLOR:
		strcpy(szMessage, "The DirectDrawSurface object is not using an 8-bit color palette and the requested operation requires an 8-bit color palette.");
		break;
	case DDERR_NOTAOVERLAYSURFACE:
		strcpy(szMessage, "An overlay component is called for a non-overlay surface.");
		break;
	case DDERR_NOTEXTUREHW:
		strcpy(szMessage, "The operation cannot be carried out because no texture-mapping hardware is present or available.");
		break;
	case DDERR_NOTFLIPPABLE:
		strcpy(szMessage, "An attempt has been made to flip a surface that cannot be flipped.");
		break;
	case DDERR_NOTFOUND:
		strcpy(szMessage, "The requested item was not found.");
		break;
	case DDERR_NOTINITIALIZED:
		strcpy(szMessage, "An attempt was made to call an interface method of a DirectDraw object created by CoCreateInstance before the object was initialized.");
		break;
	case DDERR_NOTLOADED:
		strcpy(szMessage, "The surface is an optimized surface, but it has not yet been allocated any memory");
		break;
	case DDERR_NOTLOCKED:
		strcpy(szMessage, "An attempt is made to unlock a surface that was not locked.");
		break;
	case DDERR_NOTPAGELOCKED:
		strcpy(szMessage, "An attempt is made to page unlock a surface with no outstanding page locks.");
		break;
	case DDERR_NOTPALETTIZED:
		strcpy(szMessage, "The surface being used is not a palette-based surface.");
		break;
	case DDERR_NOVSYNCHW:
		strcpy(szMessage, "The operation cannot be carried out because there is no hardware support for vertical blank synchronized operations.");
		break;
	case DDERR_NOZBUFFERHW:
		strcpy(szMessage, "The operation to create a z-buffer in display memory or to perform a blit using a z-buffer cannot be carried out because there is no hardware support for z-buffers.");
		break;
	case DDERR_NOZOVERLAYHW:
		strcpy(szMessage, "The overlay surfaces cannot be z-layered based on the z-order because the hardware does not support z-ordering of overlays.");
		break;
	case DDERR_OUTOFCAPS:
		strcpy(szMessage, "The hardware needed for the requested operation has already been allocated.");
		break;
	case DDERR_OUTOFMEMORY:
		strcpy(szMessage, "DirectDraw does not have enough memory to perform the operation.");
		break;
	case DDERR_OUTOFVIDEOMEMORY:
		strcpy(szMessage, "DirectDraw does not have enough display memory to perform the operation.");
		break;
	case DDERR_OVERLAYCANTCLIP:
		strcpy(szMessage, "The hardware does not support clipped overlays.");
		break;
	case DDERR_OVERLAYCOLORKEYONLYONEACTIVE:
		strcpy(szMessage, "An attempt was made to have more than one color key active on an overlay.");
		break;
	case DDERR_OVERLAYNOTVISIBLE:
		strcpy(szMessage, "The IDirectDrawSurface3::GetOverlayPosition method is called on a hidden overlay.");
		break;
	case DDERR_PALETTEBUSY:
		strcpy(szMessage, "Access to this palette is refused because the palette is locked by another thread.");
		break;
	case DDERR_PRIMARYSURFACEALREADYEXISTS:
		strcpy(szMessage, "This process has already created a primary surface.");
		break;
	case DDERR_REGIONTOOSMALL:
		strcpy(szMessage, "The region passed to the IDirectDrawClipper::GetClipList method is too small.");
		break;
	case DDERR_SURFACEALREADYATTACHED:
		strcpy(szMessage, "An attempt was made to attach a surface to another surface to which it is already attached.");
		break;
	case DDERR_SURFACEALREADYDEPENDENT:
		strcpy(szMessage, "An attempt was made to make a surface a dependency of another surface to which it is already dependent.");
		break;
	case DDERR_SURFACEBUSY:
		strcpy(szMessage, "Access to the surface is refused because the surface is locked by another thread.");
		break;
	case DDERR_SURFACEISOBSCURED:
		strcpy(szMessage, "Access to the surface is refused because the surface is obscured.");
		break;
	case DDERR_SURFACELOST:
		strcpy(szMessage, "Access to the surface is refused because the surface memory is gone. The DirectDrawSurface object representing this surface should have the IDirectDrawSurface3::Restore method called on it.");
		break;
	case DDERR_SURFACENOTATTACHED:
		strcpy(szMessage, "The requested surface is not attached.");
		break;
	case DDERR_TOOBIGHEIGHT:
		strcpy(szMessage, "The height requested by DirectDraw is too large.");
		break;
	case DDERR_TOOBIGSIZE:
		strcpy(szMessage, "The size requested by DirectDraw is too large. However, the individual height and width are OK.");
		break;
	case DDERR_TOOBIGWIDTH:
		strcpy(szMessage, "The width requested by DirectDraw is too large.");
		break;
	case DDERR_UNSUPPORTED:
		strcpy(szMessage, "The operation is not supported.");
		break;
	case DDERR_UNSUPPORTEDFORMAT:
		strcpy(szMessage, "The FourCC format requested is not supported by DirectDraw.");
		break;
	case DDERR_UNSUPPORTEDMASK:
		strcpy(szMessage, "The bitmask in the pixel format requested is not supported by DirectDraw.");
		break;
	case DDERR_UNSUPPORTEDMODE:
		strcpy(szMessage, "The display is currently in an unsupported mode.");
		break;
	case DDERR_VERTICALBLANKINPROGRESS:
		strcpy(szMessage, "A vertical blank is in progress.");
		break;
	case DDERR_VIDEONOTACTIVE:
		strcpy(szMessage, "The video port is not active");
		break;
	case DDERR_WASSTILLDRAWING:
		strcpy(szMessage, "The previous blit operation that is transferring information to or from this surface is incomplete.");
		break;
	case DDERR_WRONGMODE:
		strcpy(szMessage, "This surface cannot be restored because it was created in a different mode.");
		break;
	case DDERR_XALIGN:
		strcpy(szMessage, "The provided rectangle was not horizontally aligned on a required boundary.");
		break;
	default:
		strcpy(szMessage, "An unknown error has occurred.");
		break;
	}

	if(m_lpDD) m_lpDD->FlipToGDISurface();
	Release();

	char szDlgText[256];
	sprintf(szDlgText, "Fatal Exception Caught:\n%s\nFile:%s\nLine:%d", szMessage, lpszFile, nLine);
	MessageBeep(MB_ICONEXCLAMATION);
	MessageBox(NULL, szDlgText, "CDirect3D", MB_OK);

	DWORD dwNumWrit;
	HANDLE hFile = CreateFile("c:\\error.txt", GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	WriteFile(hFile, szDlgText, strlen(szDlgText), &dwNumWrit, NULL);
	CloseHandle(hFile);
	
	PostQuitMessage(1);
}

LPDIRECTDRAW CDirectDraw::GetDirectDraw()
{
	ASSERT(m_lpDD);
	return m_lpDD;
}

LPDIRECTDRAWSURFACE CDirectDraw::GetBackSurface()
{
	ASSERT(m_lpBackSurface);
	return m_lpBackSurface;
}

void CDirectDraw::Flip()
{
	HRESULT hResult;
	while(hResult = m_lpPrimarySurface->Flip(NULL, 0), hResult == DDERR_WASSTILLDRAWING);
	if(hResult == DDERR_SURFACELOST)
		VERIFY(m_lpPrimarySurface->Restore());
}

BOOL FAR PASCAL DDEnumCallback(GUID FAR* lpGUID, LPSTR lpDriverDesc, LPSTR lpDriverName, LPVOID lpContext)
{
	DDCAPS capsDriver, capsHEL;
	LPCALLBACK_PARAM pCP = (LPCALLBACK_PARAM)lpContext;

	// Try to open GUID device
	HRESULT hResult = DirectDrawCreate(lpGUID, pCP->ppDD, NULL);
	if(hResult != DD_OK) return DDENUMRET_OK; // keep looking

	// Get its caps
	memset(&capsDriver, 0, sizeof(DDCAPS));
	capsDriver.dwSize = sizeof(DDCAPS);

	memset(&capsHEL, 0, sizeof(DDCAPS));
	capsHEL.dwSize = sizeof(DDCAPS);

	hResult = (*pCP->ppDD)->GetCaps(&capsDriver, &capsHEL);
	if(hResult != DD_OK)
	{
		RELEASE(*pCP->ppDD);
		return DDENUMRET_OK; // keep looking
	}

	// If it has the caps we want, make *lppContext point to it
	if(capsDriver.dwCaps & pCP->dwDesiredCaps)
		return DDENUMRET_CANCEL; // we found one!

	// Otherwise reset it and keep looking
	RELEASE(*pCP->ppDD);
	*pCP->ppDD = NULL;

	// ...keep looking...
	return DDENUMRET_OK;
}

int CDirectDraw::Create(HWND hWnd, int nWidth, int nHeight, int nBpp, DWORD dwExtraCaps/*=0*/)
{
	ASSERT(hWnd);

	m_nBpp = nBpp;
	
	int nRet = 1; // 1 if caps success, -1 if caps failed or don't need extra caps

	// Attempt to find driver with specified capabilities...
	if(dwExtraCaps)
	{
		CALLBACK_PARAM cp;
		cp.dwDesiredCaps = dwExtraCaps;
		cp.ppDD = &m_lpDD;
		VERIFY(DirectDrawEnumerate(DDEnumCallback, &cp));
	}

	// otherwise create for the D3D HEL...
	if(!m_lpDD)
	{
		VERIFY(DirectDrawCreate(NULL, &m_lpDD, NULL));
		nRet = -1; // couldn't get caps if created for d3d
		// otherwise all is well
	}

	ASSERT(m_lpDD);
	VERIFY(m_lpDD->SetCooperativeLevel(hWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN));
	VERIFY(m_lpDD->SetDisplayMode(nWidth, nHeight, nBpp));

	// Create Primary and Back Surfaces...
	DDSURFACEDESC ddsd;
	memset(&ddsd, 0, sizeof(DDSURFACEDESC));
	ddsd.dwSize = sizeof(DDSURFACEDESC);
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP | dwExtraCaps;
	ddsd.dwBackBufferCount = 1;

	VERIFY(m_lpDD->CreateSurface(&ddsd, &m_lpPrimarySurface, NULL));

	// Retrieve Back Surface...
	DDSCAPS ddscaps;
	memset(&ddscaps, 0, sizeof(DDSCAPS));
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
	ASSERT(m_lpPrimarySurface);
	VERIFY(m_lpPrimarySurface->GetAttachedSurface(&ddscaps, &m_lpBackSurface));
	ASSERT(m_lpBackSurface);

	// Set Clipping Region...
	LPDIRECTDRAWCLIPPER lpClipper;
	VERIFY(m_lpDD->CreateClipper(0, &lpClipper, NULL));
	lpClipper->SetHWnd(0, hWnd);
	VERIFY(m_lpBackSurface->SetClipper(lpClipper));

	return nRet;
}

bool CDirectDraw::Blit(LPRECT pDestRect, CBmpFile &bmp)
{
	LPDIRECTDRAWSURFACE pSrcSurface = CreateSurface(bmp);
	bool bRet = Blit(pDestRect, pSrcSurface);
	RELEASE(pSrcSurface);
	return bRet;
}

bool CDirectDraw::Blit(LPRECT pDestRect, LPDIRECTDRAWSURFACE pSurface)
{
	HRESULT hResult;

	ASSERT(m_lpBackSurface);
	hResult = m_lpBackSurface->Blt(pDestRect, pSurface, NULL, DDBLT_KEYSRC | DDBLT_WAIT | DDBLT_ASYNC, NULL);
	if(hResult == DDERR_SURFACELOST)
	{
		VERIFY(pSurface->Restore());
		return true;
	}

	return true;
}

LPDIRECTDRAWSURFACE CDirectDraw::CreateSurface(CBmpFile &bmp)
{
	// This function should convert the bmp, whatever format it
	// is in, over to a surface compatible with this Direct
	// Draw object.

	int nWidth = bmp.Width(),
		nHeight = bmp.Height();

	LPDIRECTDRAWSURFACE pSurface =
		CreateSurface(nWidth, nHeight);
	if(!pSurface) return NULL;

	HBITMAP hBmp = bmp.CreateHandle(NULL);
	if(!hBmp) return NULL;

	HDC hSurfaceDC;
	if(pSurface->GetDC(&hSurfaceDC) != DD_OK)
		return NULL;

	HDC hBmpDC = CreateCompatibleDC(hSurfaceDC);
	if(!hBmpDC) return NULL;

	HGDIOBJ pOldObj = SelectObject(hBmpDC, hBmp);

	BitBlt(hSurfaceDC, 0, 0, bmp.Width(), bmp.Height(), hBmpDC, 0, 0, SRCCOPY);

	pSurface->ReleaseDC(hSurfaceDC);
	DeleteDC(hBmpDC);
	DeleteObject(hBmp);

	return pSurface;
}

LPDIRECTDRAWSURFACE CDirectDraw::CreateSurface(int nWidth, int nHeight)
{
	LPDIRECTDRAWSURFACE pSurface = NULL;

	// Describe Surface...
	DDSURFACEDESC ddsd;
	memset(&ddsd, 0, sizeof(DDSURFACEDESC));
	ddsd.dwSize = sizeof(DDSURFACEDESC);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwWidth = nWidth; // in pixels
	ddsd.dwHeight = nHeight;

	// Create...
	VERIFY(m_lpDD->CreateSurface(&ddsd, &pSurface, NULL));
	if(!pSurface) return NULL;

	// Set Transparant Color (black)...
	DDCOLORKEY ddck;
    ddck.dwColorSpaceLowValue = 0;
    ddck.dwColorSpaceHighValue = 0;
    VERIFY(pSurface->SetColorKey(DDCKEY_SRCBLT, &ddck));

	return pSurface;
}

void CDirectDraw::BlitToSurface(LPDIRECTDRAWSURFACE pDestSurface, LPBITMAPINFO pBmpInfo)
{
	ASSERT(pDestSurface);

	LPBITMAPINFOHEADER pBmpInfoHeader = (LPBITMAPINFOHEADER)pBmpInfo;

	HDC hdc;
	ASSERT(pDestSurface->GetDC(&hdc) == DD_OK);

	LPBYTE pBmpBits = ((LPBYTE)pBmpInfo) + pBmpInfo->bmiHeader.biSize;

	LPBYTE pGDIBits;
	HBITMAP hBmp = CreateDIBSection(hdc, pBmpInfo, DIB_RGB_COLORS, (LPVOID*)&pGDIBits, NULL, 0);
	ASSERT(hBmp);

	memcpy(pGDIBits, pBmpBits, pBmpInfoHeader->biSizeImage);

	HDC hBmpDC = CreateCompatibleDC(hdc);
	ASSERT(hBmpDC);

	HGDIOBJ pOldObj = SelectObject(hBmpDC, hBmp);
	BitBlt(hdc, 0, 0, pBmpInfoHeader->biWidth, pBmpInfoHeader->biHeight, hBmpDC, 0, 0, SRCCOPY);

	if(pOldObj) SelectObject(hBmpDC, pOldObj);
	DeleteObject(hBmp);

	DeleteDC(hBmpDC);

	ASSERT(pDestSurface->ReleaseDC(hdc) == DD_OK);
}

LPVOID CDirectDraw::LockSurface(LPDIRECTDRAWSURFACE lpSurface/*=NULL*/, LPDDSURFACEDESC lpDesc/*=NULL*/)
{
	if(lpSurface==NULL)
		lpSurface = m_lpBackSurface;

	ASSERT(lpSurface);

	// Lock the surface...
	DDSURFACEDESC ddsd;
	if(!lpDesc) lpDesc = &ddsd;

	memset(lpDesc, 0, sizeof(DDSURFACEDESC));
	lpDesc->dwSize = sizeof(DDSURFACEDESC);
	VERIFY(lpSurface->Lock(NULL, lpDesc, DDLOCK_WAIT, NULL));
	return lpDesc->lpSurface;

	// Surface must be unlocked using UnlockSurface();
}

void CDirectDraw::UnlockSurface(LPVOID lpData, LPDIRECTDRAWSURFACE lpSurface/*=NULL*/)
{
	// Unlock...
	if(!lpSurface) lpSurface = m_lpBackSurface;

	ASSERT(lpSurface);
	ASSERT(lpData);
	VERIFY(lpSurface->Unlock(lpData));
}

int CDirectDraw::GetBpp()
{
	return m_nBpp;
}

bool CDirectDraw::GetDC(HDC *pHDC)
{
	ASSERT(pHDC);
	HRESULT hResult = m_lpBackSurface->GetDC(pHDC);
	if(hResult != DD_OK)
	{
		ProcessResultA(hResult, __FILE__, __LINE__);
		return false;
	}

	return true;
}

bool CDirectDraw::ReleaseDC(HDC *pHDC)
{
	ASSERT(pHDC);
	HRESULT hResult = m_lpBackSurface->ReleaseDC(*pHDC);
	if(hResult != DD_OK)
	{
		ProcessResultA(hResult, __FILE__, __LINE__);
		return false;
	}

	pHDC = NULL;
	return true;
}

bool CDirectDraw::CreateZBuffer(bool bHardware, int nWidth, int nHeight)
{
	ASSERT(m_lpDD);
	ASSERT(m_lpBackSurface);

	// (z-buffer surface description)
	DDSURFACEDESC ddsd;
	memset(&ddsd, 0, sizeof(DDSURFACEDESC));
	ddsd.dwSize = sizeof(DDSURFACEDESC);

	// (this depends on hardware or software mode)
	DWORD dwExtraCaps;

	// First need to get dd's caps for z-buffers
	// to find best settings for hardware or software
	DDCAPS capsDriver, capsHEL;
	memset(&capsDriver, 0, sizeof(DDCAPS));
	capsDriver.dwSize = sizeof(DDCAPS);

	memset(&capsHEL, 0, sizeof(DDCAPS));
	capsHEL.dwSize = sizeof(DDCAPS);

	HRESULT hResult = m_lpDD->GetCaps(&capsDriver, &capsHEL);
	if(hResult != DD_OK)
	{
		Release();
		return false;
	}

	if(bHardware)
	{
		if(capsDriver.dwZBufferBitDepths & DDBD_16)
			ddsd.dwZBufferBitDepth = 16;
		else if(capsDriver.dwZBufferBitDepths & DDBD_8)
			ddsd.dwZBufferBitDepth = 8;
		else return false;

		dwExtraCaps = DDSCAPS_VIDEOMEMORY;
	}
	else // HEL
	{
		if(capsHEL.dwZBufferBitDepths & DDBD_16)
			ddsd.dwZBufferBitDepth = 16;
		else if(capsHEL.dwZBufferBitDepths & DDBD_8)
			ddsd.dwZBufferBitDepth = 8;
		else return false;

		dwExtraCaps = DDSCAPS_SYSTEMMEMORY;
	}

	// Create ZBuffer...
	ddsd.dwFlags = DDSD_WIDTH | DDSD_HEIGHT | DDSD_CAPS | DDSD_ZBUFFERBITDEPTH;
	ddsd.dwWidth = nWidth;
	ddsd.dwHeight = nHeight;
	ddsd.ddsCaps.dwCaps = DDSCAPS_ZBUFFER | dwExtraCaps;

	VERIFY(m_lpDD->CreateSurface(&ddsd, &m_lpZSurface, NULL));
	VERIFY(m_lpBackSurface->AddAttachedSurface(m_lpZSurface));

	return true;
}

void CDirectDraw::ClearSurface(LPDIRECTDRAWSURFACE pSurface)
{
	DDSURFACEDESC desc;
	LPBYTE p = (LPBYTE)LockSurface(pSurface, &desc);

	int nH = desc.dwHeight,
		nW = desc.dwWidth * GetBpp()>>3,
		nTW = desc.lPitch;

	while(nH--)
	{
		ZeroMemory(p, nW);
		p += nTW;
	}

	UnlockSurface(p, pSurface);

	return;
}
